--[[
------
|TODO|
------

]]

-- settings

local settings = {
	barLength = 200,
	selfBarLength = 320,
	smallBarLength = 100,
	showBackground = false,
	barWidth = 6,
	selfBarWidth = 6,
	smallBarWidth = 6,
	lowManaThreshold = 25, -- In percent
	lowHealthThreshold = 25, -- In percent
	interval = 0.05,
	intervalShow = 20,
	version = 2,
	customBars = {},
	expansionMode = false,
}

local expansionVersion = 40000

--[[function tuttiHUDPrintSettings()
	-- Used for debugging.
	for k,v in pairs(settings) do
		print(k .. ": " .. tostring(v))
	end
end]]

local function expansionMode()
	local _, _, _, gameVersion = GetBuildInfo()
	return (gameVersion == expansionVersion) or settings.expansionMode
end

local function updateSettings(oldSettings)
	-- Returns newSettings: a settings array that's compatible with the current version,
	-- with old settings imported where possible.

	-- Function is vestigial for now, as only one settings version exists.
	local newSettings
	if oldSettings.version == 0 then -- Version number 0 is for testing.
		newSettings = oldSettings
	end
	if oldSettings.version == 1 then
		-- When a new version is released, the update code goes here
		oldSettings.version = 2
		oldSettings.expansionMode = false
		newSettings = oldSettings
	end

	return newSettings
end

local targets = {
	"player",
	"target",
	"focus",
	"targettarget",
	"focustarget",
	"pet",
}

local timePassed = 0

local function makeTexture(f, type)
	local r = f:CreateTexture()
	r:SetTexture("Interface\\AddOns\\tuttiHUD\\textures\\" .. type .. "bar.blp")
	r:SetAllPoints(f)
	f.texture = r
end

local function makeColourTexture(f, c)
	local r = f:CreateTexture()
	r:SetTexture(c[1], c[2], c[3])
	r:SetAllPoints(f)
	f.texture = r
end

local function changeTexture(f, type)
	local r = f.texture
	r:SetTexture("Interface\\AddOns\\tuttiHUD\\textures\\" .. type .. "bar.blp")
	r:SetAllPoints(f)
	f.texture = r
end

local function removeTexture(f)
	if f.texture then
		f.texture:SetTexture(nil)
		f.texture = nil
	end
end

-- Set up all the bars

local player = CreateFrame("Frame", nil, UIParent, "tuttiHUDBarTemplateH")
local target = CreateFrame("Frame", nil, UIParent, "tuttiHUDBarTemplateV")
local focus = CreateFrame("Frame", nil, UIParent, "tuttiHUDBarTemplateV")
target.target = CreateFrame("Frame", nil, UIParent, "tuttiHUDBarTemplateV")
focus.target = CreateFrame("Frame", nil, UIParent, "tuttiHUDBarTemplateV")
player.more = CreateFrame("Frame", nil, UIParent, "tuttiHUDBarTemplateH")
player.pet = CreateFrame("Frame", nil, UIParent, "tuttiHUDBarTemplateH")

player.health = CreateFrame("Frame", nil, player, "tuttiHUDBarTemplateH")
player.resource = CreateFrame("Frame", nil, player, "tuttiHUDBarTemplateH")
player.mana = CreateFrame("Frame", nil, player, "tuttiHUDBarTemplateH")
target.health = CreateFrame("Frame", nil, target, "tuttiHUDBarTemplateV")
target.resource = CreateFrame("Frame", nil, target, "tuttiHUDBarTemplateV")
focus.health = CreateFrame("Frame", nil, focus, "tuttiHUDBarTemplateV")
focus.resource = CreateFrame("Frame", nil, focus, "tuttiHUDBarTemplateV")
target.target.health = CreateFrame("Frame", nil, target.target, "tuttiHUDBarTemplateV")
target.target.resource = CreateFrame("Frame", nil, target.target, "tuttiHUDBarTemplateV")
focus.target.health = CreateFrame("Frame", nil, focus.target, "tuttiHUDBarTemplateV")
focus.target.resource = CreateFrame("Frame", nil, focus.target, "tuttiHUDBarTemplateV")
player.more.mana = CreateFrame("Frame", nil, player.more, "tuttiHUDBarTemplateH")
if expansionMode() then
	player.more.eclipse1 = CreateFrame("Frame", nil, player.more, "tuttiHUDBarTemplateH")
	player.more.eclipse2 = CreateFrame("Frame", nil, player.more, "tuttiHUDBarTemplateH")
end
player.more.blood1 = CreateFrame("Frame", nil, player.more, "tuttiHUDBarTemplateH")
player.more.blood2 = CreateFrame("Frame", nil, player.more, "tuttiHUDBarTemplateH")
player.more.frost1 = CreateFrame("Frame", nil, player.more, "tuttiHUDBarTemplateH")
player.more.frost2 = CreateFrame("Frame", nil, player.more, "tuttiHUDBarTemplateH")
player.more.unholy1 = CreateFrame("Frame", nil, player.more, "tuttiHUDBarTemplateH")
player.more.unholy2 = CreateFrame("Frame", nil, player.more, "tuttiHUDBarTemplateH")
player.pet.health = CreateFrame("Frame", nil, player.pet, "tuttiHUDBarTemplateH")
player.pet.resource = CreateFrame("Frame", nil, player.pet, "tuttiHUDBarTemplateH")

local powerTypes = {
	[0] = "mana",
	[1] = "rage",
	[2] = "focus", -- NYI
	[3] = "energy",
	[4] = "",
	[5] = "runes",
	[6] = "runic",
}

local function druidIsBalance()
	-- Will actually return true for any class using spec #1, not just druids
	-- Only works right for Cataclysm! Before Cataclysm having more than 10 points in
	-- a different tree is allowed.
	local _, _, _, _, points = GetTalentTabInfo(1)
	if points > 10 then return true end
	if points > 0 and UnitLevel("player") < 70 then return true end
	return false
end

-- Bottom: Player's health and resource (and mana/eclipse/runes if applicable, NYI)

makeTexture(player.health, "health")
player.health:SetPoint("TOPLEFT")
makeTexture(player.resource, "mana") -- The power type doesn't load immediately upon login, change it later
player.resource:SetPoint("BOTTOMLEFT")

player.more:SetPoint("TOPLEFT", player, "BOTTOMLEFT")

local _, class = UnitClass("player")

if class == "DRUID" then makeTexture(player.more.mana, "mana")
elseif class == "WARLOCK" and expansionMode() then makeTexture(player.more.mana, "soulshard") -- Change texture
elseif class == "PALADIN" and expansionMode() then makeTexture(player.more.mana, "sun") -- Just needs to be light
else makeTexture(player.more.mana, "no") end -- To prevent issues
player.more.mana:SetPoint("TOPLEFT")

if expansionMode() then
	makeTexture(player.more.eclipse1, "nomoon")
	player.more.eclipse1:SetPoint("BOTTOMRIGHT", player.more, "BOTTOM")
	makeTexture(player.more.eclipse2, "nosun")
	player.more.eclipse2:SetPoint("BOTTOMLEFT", player.more, "BOTTOM")
end

makeTexture(player.more.blood1, "low") -- The colours are too similar to see a difference
makeTexture(player.more.blood2, "low")
makeTexture(player.more.frost1, "frost")
makeTexture(player.more.frost2, "frost")
makeTexture(player.more.unholy1, "unholy")
makeTexture(player.more.unholy2, "unholy")
player.more.blood1:SetPoint("TOPLEFT") -- These two always go in these corners, the other four
player.more.blood2:SetPoint("BOTTOMLEFT") -- are set to different locations in saveSettings().
local runebars = {player.more.blood1, player.more.blood2, player.more.unholy1, player.more.unholy2, player.more.frost1, player.more.frost2}

player.more.mana:Hide()
if expansionMode() then
	player.more.eclipse1:Hide()
	player.more.eclipse2:Hide()
end
player.more.blood1:Hide()
player.more.blood2:Hide()
player.more.frost1:Hide()
player.more.frost2:Hide()
player.more.unholy1:Hide()
player.more.unholy2:Hide()
if class == "DEATHKNIGHT" then
	player.more.blood1:Show()
	player.more.blood2:Show()
	player.more.frost1:Show()
	player.more.frost2:Show()
	player.more.unholy1:Show()
	player.more.unholy2:Show()
elseif class == "DRUID" then
	player.more.mana:Show()
	if expansionMode() then
		player.more.eclipse1:Show()
		player.more.eclipse2:Show()
	end
elseif class == "WARLOCK" and expansionMode() then
	player.more.mana:Show()
elseif class == "PALADIN" and expansionMode() then
	player.more.mana:Show()
else player.more:Hide() end

-- Below that: Pet's health and resource

powerType = UnitPowerType("pet")

makeTexture(player.pet.health, "health")
player.pet.health:SetPoint("TOPLEFT")
makeTexture(player.pet.resource, powerTypes[powerType])
player.pet.resource:SetPoint("BOTTOMLEFT")

player.pet:Hide()

-- Left: Target's health and resource

powerType = UnitPowerType("target")

makeTexture(target.health, "health")
target.health:SetPoint("BOTTOMLEFT")
makeTexture(target.resource, powerTypes[powerType])
target.resource:SetPoint("BOTTOMRIGHT")

target:Hide()

-- Further left: Target's target's health and resource

powerType = UnitPowerType("targettarget")

makeTexture(target.target.health, "health")
target.target.health:SetPoint("BOTTOMLEFT")
makeTexture(target.target.resource, powerTypes[powerType])
target.target.resource:SetPoint("BOTTOMRIGHT")

target.target:Hide()

-- Right: Focus' health and resource

powerType = UnitPowerType("focus")

makeTexture(focus.health, "health")
focus.health:SetPoint("BOTTOMLEFT")
makeTexture(focus.resource, powerTypes[powerType])
focus.resource:SetPoint("BOTTOMRIGHT")

focus:Hide()

-- Further right: Focus' target's health and resource

powerType = UnitPowerType("focustarget")

makeTexture(focus.target.health, "health")
focus.target.health:SetPoint("BOTTOMLEFT")
makeTexture(focus.target.resource, powerTypes[powerType])
focus.target.resource:SetPoint("BOTTOMRIGHT")

focus.target:Hide()

-- Event hooking

local eventFrame = CreateFrame("Frame")
--eventFrame:RegisterEvent("UNIT_HEALTH")
--eventFrame:RegisterEvent("UNIT_MANA")
--eventFrame:RegisterEvent("UNIT_ENERGY")
--eventFrame:RegisterEvent("UNIT_RAGE")
--eventFrame:RegisterEvent("UNIT_RUNIC_POWER")
--eventFrame:RegisterEvent("UNIT_DISPLAYPOWER")
--eventFrame:RegisterEvent("PLAYER_FOCUS_CHANGED")
eventFrame:RegisterEvent("ADDON_LOADED")
--eventFrame:RegisterEvent("UNIT_TARGET")
eventFrame:RegisterEvent("PLAYER_LOGOUT")

local function saveSettings()
	player:SetWidth(settings.selfBarLength)
	player:SetHeight(settings.selfBarWidth * 2)
	target:SetWidth(settings.barWidth * 2)
	target:SetHeight(settings.barLength)
	focus:SetWidth(settings.barWidth * 2)
	focus:SetHeight(settings.barLength)
	target.target:SetWidth(settings.smallBarWidth * 2)
	target.target:SetHeight(settings.smallBarLength)
	focus.target:SetWidth(settings.smallBarWidth * 2)
	focus.target:SetHeight(settings.smallBarLength)
	player.pet:SetWidth(settings.smallBarLength)
	player.pet:SetHeight(settings.smallBarWidth * 2)
	player.more:SetWidth(settings.selfBarLength)
	local size, class = UnitClass("player") -- Save a line for declaration of size
	if class == "DEATHKNIGHT" then
		size = 2
	elseif class == "DRUID" or ((class == "WARLOCK" or class == "PALADIN") and expansionMode()) then
		size = 1
	else
		size = 0
	end
	player.more:SetHeight(size * settings.selfBarWidth)
	player.more.frost1:SetPoint("TOPLEFT", settings.selfBarLength / 3, 0)
	player.more.frost2:SetPoint("BOTTOMLEFT", settings.selfBarLength / 3, 0)
	player.more.unholy1:SetPoint("TOPLEFT", 2 * settings.selfBarLength / 3, 0)
	player.more.unholy2:SetPoint("BOTTOMLEFT", 2 * settings.selfBarLength / 3, 0)
	player:SetPoint("CENTER", 0, -((settings.barLength / 2) + settings.selfBarWidth))
	target:SetPoint("CENTER", -((settings.selfBarLength / 2) + settings.barWidth), 0)
	target.target:SetPoint("CENTER", -((settings.selfBarLength / 2) + (settings.barWidth * 2) + (settings.smallBarWidth * 2)), 0)
	focus:SetPoint("CENTER", ((settings.selfBarLength / 2) + settings.barWidth), 0)
	focus.target:SetPoint("CENTER", ((settings.selfBarLength / 2) + (settings.barWidth * 2) + (settings.smallBarWidth * 2)), 0)
	player.pet:SetPoint("CENTER", 0, -((settings.barLength / 2) + (settings.selfBarWidth * (size + 2)) + (settings.smallBarWidth * 2)))
	player.health:SetHeight(settings.selfBarWidth)
	player.resource:SetHeight(settings.selfBarWidth)
	target.health:SetWidth(settings.barWidth)
	target.resource:SetWidth(settings.barWidth)
	focus.health:SetWidth(settings.barWidth)
	focus.resource:SetWidth(settings.barWidth)
	target.target.health:SetWidth(settings.smallBarWidth)
	target.target.resource:SetWidth(settings.smallBarWidth)
	focus.target.health:SetWidth(settings.smallBarWidth)
	focus.target.resource:SetWidth(settings.smallBarWidth)
	player.more.mana:SetHeight(settings.selfBarWidth)
	if expansionMode() then
		player.more.eclipse1:SetHeight(settings.selfBarWidth)
		player.more.eclipse2:SetHeight(settings.selfBarWidth)
	end
	player.more.blood1:SetHeight(settings.selfBarWidth)
	player.more.blood2:SetHeight(settings.selfBarWidth)
	player.more.frost1:SetHeight(settings.selfBarWidth)
	player.more.frost2:SetHeight(settings.selfBarWidth)
	player.more.unholy1:SetHeight(settings.selfBarWidth)
	player.more.unholy2:SetHeight(settings.selfBarWidth)
	if settings.showBackground then
		-- TODO: Detect if a texture exists and change it instead of creating a new one if it does
		makeTexture(player, "no")
		makeTexture(target, "no")
		makeTexture(focus, "no")
		makeTexture(target.target, "no")
		makeTexture(focus.target, "no")
		makeTexture(player.more, "no")
		makeTexture(player.pet, "no")
	else
		removeTexture(player)
		removeTexture(target)
		removeTexture(focus)
		removeTexture(target.target)
		removeTexture(focus.target)
		removeTexture(player.more)
		removeTexture(player.pet)
	end

	settings.interval = 1 / (settings.intervalShow * 1.0)
end

local function setUpBar(i)
	local c = settings.customBars
	c[i].background = CreateFrame("Frame", c[i].name, UIParent, "tuttiHUDBarTemplate" .. (c[i].rotation and "V" or "H"))
	makeTexture(c[i].background, "no")
	if c[i].rotation then
		c[i].background:SetWidth(c[i].width)
		c[i].background:SetHeight(c[i].length)
	else
		c[i].background:SetWidth(c[i].length)
		c[i].background:SetHeight(c[i].width)
	end
	c[i].background:SetPoint("CENTER")
	c[i].bar = CreateFrame("Frame", c[i].name .. "child", c[i].background, "tuttiHUDBarTemplate" .. (c[i].rotation and "V" or "H"))
	makeColourTexture(c[i].bar, c[i].colour)
	c[i].bar:SetPoint("BOTTOMLEFT")
	if c[i].rotation then
		c[i].bar:SetWidth(c[i].width)
		c[i].bar:SetHeight(c[i].length)
	else
		c[i].bar:SetWidth(c[i].length)
		c[i].bar:SetHeight(c[i].width)
	end
	c[i].background:EnableMouse(true)
	c[i].background:SetMovable(true)
	c[i].background:RegisterForDrag("LeftButton")
	c[i].background:SetScript("OnDragStart", c[i].background.StartMoving)
	c[i].background:SetScript("OnDragStop", c[i].background.StopMovingOrSizing)
end

local function onEvent(frame, event, ...)
	if event == "ADDON_LOADED" then
		if select(1, ...) == "tuttiHUD" then
			-- Even at this point UnitPowerType() doesn't return the right function... workaround time.
			powerType, class = UnitClass("player")
			if (class == "WARRIOR" or UnitBuff("player", "Bear Form") or UnitBuff("player", "Dire Bear Form")) then powerType = "rage"
			elseif (class == "ROGUE" or UnitBuff("player", "Cat Form")) then powerType = "energy"
			elseif class == "DEATHKNIGHT" then powerType = "runic" end
			changeTexture(player.resource, powerType)
			onEvent(eventFrame, "UNIT_RAGE", "player")
			if tuttiHUDSettings and settings.version ~= 0 then -- Version number 0 is for testing
				if tuttiHUDSettings.version == settings.version then
					-- Everything is up to date, load saved settings.
					settings = tuttiHUDSettings
				else
					-- New settings version - just reset for now.
					tuttiHUDSettings = updateSettings(tuttiHUDSettings)
					settings = tuttiHUDSettings
				end
			else
				-- No saved variables, save the default ones.
				tuttiHUDSettings = settings
			end
			saveSettings()
			for i=1, #settings.customBars do
				setUpBar(i)
			end
		end
	end
	if event == "PLAYER_LOGOUT" then
		tuttiHUDSettings = settings
	end
end

local function onUpdate(_, elapsed)
	timePassed = timePassed + elapsed
	if timePassed > settings.interval then
		-- Keep everything up to date

		local _, class = UnitClass("player")
		local cur, max, percent, barLength
		-- Player
		-- Health
		cur, max = UnitHealth("player"), UnitHealthMax("player")
		percent = (cur / max) * 100
		barLength = (cur / max) * settings.selfBarLength
		player.health:SetWidth(barLength)
		if percent <= settings.lowHealthThreshold then
			changeTexture(player.health, "low")
		else
			changeTexture(player.health, "health")
		end
		if barLength == 0 then player.health:Hide() else player.health:Show() end
		-- Mana/resource
		local powerType = UnitPowerType("player")
		changeTexture(player.resource, powerTypes[powerType])
		cur, max = UnitPower("player"), UnitPowerMax("player")
		percent, barLength = (cur / max) * 100, (cur / max) * settings.selfBarLength
		player.resource:SetWidth(barLength)
		if UnitPowerType("player") == 0 then
			if percent <= settings.lowManaThreshold then
				changeTexture(player.resource, "low")
			else
				changeTexture(player.resource, "mana")
			end
		end
		if barLength == 0 then player.resource:Hide() else player.resource:Show() end
		-- Mana (druids in non-mana forms) or eclipse (druids in moonkin form)
		if class == "DRUID" then
			player.more:Show()
			if not (UnitPowerType("player") == 0) then
				player.more.mana:Show()
				if expansionMode() then
					player.more.eclipse1:Hide()
					player.more.eclipse2:Hide()
				end
				cur = UnitPower("player", 0)
				max = UnitPowerMax("player", 0)
				percent = (cur / max) * 100
				barLength = (cur / max) * settings.selfBarLength
				if percent <= settings.lowManaThreshold then
					changeTexture(player.more.mana, "low")
				else
					changeTexture(player.more.mana, "mana")
				end
				player.more.mana:SetWidth(barLength)
			elseif druidIsBalance() and expansionMode() then
				player.more.mana:Hide()
				cur = UnitPower("player", 8)
				if cur < 0 then
					player.more.eclipse1:Show()
					player.more.eclipse1:SetWidth((cur / -100) * (settings.selfBarLength / 2))
					player.more.eclipse2:Hide()
					if UnitBuff("player", "Eclipse (Lunar)") then
						changeTexture(player.more.eclipse1, "moon")
					else
						changeTexture(player.more.eclipse1, "nomoon")
					end
				elseif cur > 0 then
					player.more.eclipse2:Show()
					player.more.eclipse2:SetWidth((cur / 100) * (settings.selfBarLength / 2))
					player.more.eclipse1:Hide()
					if UnitBuff("player", "Eclipse (Solar)") then
						changeTexture(player.more.eclipse2, "sun")
					else
						changeTexture(player.more.eclipse2, "nosun")
					end
				else
					player.more.eclipse1:Hide()
					player.more.eclipse2:Hide()
				end
			else
				player.more:Hide()
			end
		end
		local curHealth, maxHealth, curRes, maxRes, ratioHealth, ratioRes
		for i=2, 6 do
			curHealth, maxHealth = UnitHealth(targets[i]), UnitHealthMax(targets[i])
			curRes, maxRes = UnitPower(targets[i]), UnitPowerMax(targets[i])
			ratioHealth = (curHealth / maxHealth)
			ratioRes = (curRes / maxRes)
			if UnitExists(targets[i]) and i == 2 then
				local powerType = UnitPowerType("target")
				changeTexture(target.resource, powerTypes[powerType])
				barLength = ratioHealth * settings.barLength
				target.health:SetHeight(barLength)
				if barLength == 0 then target.health:Hide() else target.health:Show() end
				barLength = ratioRes * settings.barLength
				target.resource:SetHeight(barLength)
				if barLength == 0 then target.resource:Hide() else target.resource:Show() end
			elseif UnitExists(targets[i]) and i == 3 then
				local powerType = UnitPowerType("focus")
				changeTexture(focus.resource, powerTypes[powerType])
				barLength = ratioHealth * settings.barLength
				focus.health:SetHeight(barLength)
				if barLength == 0 then focus.health:Hide() else focus.health:Show() end
				barLength = ratioRes * settings.barLength
				focus.resource:SetHeight(barLength)
				if barLength == 0 then focus.resource:Hide() else focus.resource:Show() end
			elseif UnitExists(targets[i]) and i == 4 then
				local powerType = UnitPowerType("targettarget")
				changeTexture(target.target.resource, powerTypes[powerType])
				barLength = ratioHealth * settings.smallBarLength
				target.target.health:SetHeight(barLength)
				if barLength == 0 then target.target.health:Hide() else target.target.health:Show() end
				barLength = ratioRes * settings.smallBarLength
				target.target.resource:SetHeight(barLength)
				if barLength == 0 then target.target.resource:Hide() else target.target.resource:Show() end
			elseif UnitExists(targets[i]) and i == 5 then
				local powerType = UnitPowerType("focustarget")
				changeTexture(focus.target.resource, powerTypes[powerType])
				barLength = ratioHealth * settings.smallBarLength
				focus.target.health:SetHeight(barLength)
				if barLength == 0 then focus.target.health:Hide() else focus.target.health:Show() end
				barLength = ratioRes * settings.smallBarLength
				focus.target.resource:SetHeight(barLength)
				if barLength == 0 then focus.target.resource:Hide() else focus.target.resource:Show() end
			elseif UnitExists(targets[i]) and i == 6 then
				local powerType = UnitPowerType("pet")
				changeTexture(player.pet.resource, powerTypes[powerType])
				barLength = ratioHealth * settings.smallBarLength
				player.pet.health:SetWidth(barLength)
				if barLength == 0 then player.pet.health:Hide() else player.pet.health:Show() end
				barLength = ratioRes * settings.smallBarLength
				player.pet.resource:SetWidth(barLength)
				if barLength == 0 then player.pet.resource:Hide() else player.pet.resource:Show() end
			end
		end
		if UnitExists("target") then target:Show() else target:Hide() end
		if UnitExists("focus") then focus:Show() else focus:Hide() end
		if UnitExists("targettarget") then target.target:Show() else target.target:Hide() end
		if UnitExists("focustarget") then focus.target:Show() else focus.target:Hide() end
		if UnitExists("pet") then player.pet:Show() else player.pet:Hide() end

		-- Death knight runes
		if class == "DEATHKNIGHT" then
			local start, duration, ready, passed, length = 0, 0, 0, 0, 0
			for i=1, 6 do
				start, duration, ready = GetRuneCooldown(i)
				passed = GetTime() - start
				if ready then length = 1 else length = (passed / duration) end
				runebars[i]:SetWidth(length * (settings.selfBarLength / 3))
			end
		end

		-- Warlock soul shards, paladin holy power
		if (class == "WARLOCK" or class == "PALADIN") and expansionMode() then
			cur = UnitPower("player", class == "WARLOCK" and 7 or 9)
			if cur == 0 then
				player.more.mana:Hide()
			else
				player.more.mana:Show()
				player.more.mana:SetWidth((cur / 3) * settings.selfBarLength)
			end
		end

		-- And the custom bars
		local c = settings.customBars
		for i=1, #c do
			cur = assert(loadstring(c[i].func))()
			max = assert(loadstring(c[i].max))()
			local l = (cur / max) * c[i].length
			if c[i].rotation then c[i].bar:SetHeight(l) else c[i].bar:SetWidth(l) end
		end

		timePassed = timePassed - settings.interval
	end
end

eventFrame:SetScript("OnEvent", onEvent)
eventFrame:SetScript("OnUpdate", onUpdate)

-- Settings UI

local optionsFrameString = "frame|TuttiHUD Options frame" ..
	"|CENTER" ..
		"&text:Options:200:22" ..
	"|LEFT" ..
		"&text:Bars:200:18" ..
		"&text:Other:100:18" ..
	"|LEFT" ..
		"&text:Player:400:14" ..
	"|LEFT" ..
		"&text:Length:64:10" ..
		"&text:Width:136:10" ..
		"&text:Low health threshold(%):150:10" ..
	"|LEFT" ..
		"&editbox:64:100" ..
		"&editbox:64:100" ..
		"&space:68:24" ..
		"&editbox:64:100" ..
	"|LEFT" ..
		"&text:Target and focus:400:14" ..
	"|LEFT" ..
		"&text:Length:64:10" ..
		"&text:Width:136:10" ..
		"&text:Low mana threshold(%):150:10" ..
	"|LEFT" ..
		"&editbox:64:100" ..
		"&editbox:64:100" ..
		"&space:68:24" ..
		"&editbox:64:100" ..
	"|LEFT" ..
		"&text:Target of target/focus:400:14" ..
	"|LEFT" ..
		"&text:Length:64:10" ..
		"&text:Width:136:10" ..
		"&text:Update interval:150:10" ..
	"|LEFT" ..
		"&editbox:64:100" ..
		"&editbox:64:100" ..
		"&space:68:24" ..
		"&editbox:64:100" ..
	"|LEFT" ..
		"&checkbox:300:Show bar background" ..
	"|CENTER" ..
		"&button:Save" ..
		"&button:Cancel"

local optionsFrame = TuttiFrameSetupParse(optionsFrameString)
for i=1,9 do optionsFrame.editbox[i]:SetNumeric(true) end
optionsFrame:SetPoint("CENTER")
optionsFrame:Hide()

local function showOptions()
	optionsFrame.editbox[1]:SetText(settings.selfBarLength)
	optionsFrame.editbox[2]:SetText(settings.selfBarWidth)
	optionsFrame.editbox[4]:SetText(settings.barLength)
	optionsFrame.editbox[5]:SetText(settings.barWidth)
	optionsFrame.editbox[7]:SetText(settings.smallBarLength)
	optionsFrame.editbox[8]:SetText(settings.smallBarWidth)
	optionsFrame.editbox[3]:SetText(settings.lowHealthThreshold)
	optionsFrame.editbox[6]:SetText(settings.lowManaThreshold)
	optionsFrame.editbox[9]:SetText(settings.intervalShow)
	optionsFrame.checkbox[1]:SetChecked(settings.showBackground)
	optionsFrame:Show()
end

local function cancelOptions()
	optionsFrame:Hide()
end

local function saveOptions()
	settings.selfBarLength = tonumber(optionsFrame.editbox[1]:GetText())
	settings.selfBarWidth = tonumber(optionsFrame.editbox[2]:GetText())
	settings.barLength = tonumber(optionsFrame.editbox[4]:GetText())
	settings.barWidth = tonumber(optionsFrame.editbox[5]:GetText())
	settings.smallBarLength = tonumber(optionsFrame.editbox[7]:GetText())
	settings.smallBarWidth = tonumber(optionsFrame.editbox[8]:GetText())
	settings.lowHealthThreshold = tonumber(optionsFrame.editbox[3]:GetText())
	settings.lowManaThreshold = tonumber(optionsFrame.editbox[6]:GetText())
	settings.intervalShow = tonumber(optionsFrame.editbox[9]:GetText())
	settings.showBackground = optionsFrame.checkbox[1]:GetChecked()
	saveSettings()
	optionsFrame:Hide()
end

optionsFrame.button[1]:SetScript("OnClick", saveOptions)
optionsFrame.button[2]:SetScript("OnClick", cancelOptions)

-- Custom bars

local customBarRotations = {
	["VERTICAL"] = true,
	["HORIZONTAL"] = false,
}

local function createBar(r, g, b, l, w, ro, m, f, n)
	settings.customBars[#settings.customBars + 1] = {colour = {r, g, b}, length = l, width = w, rotation = customBarRotations[ro], max = m, func = f, name = n or "bar" .. GetTime()}
	saveSettings()
	setUpBar(#settings.customBars)
end

local function removeBar(ID)
	settings.customBars[ID].background:Hide()
	settings.customBars[ID].bar = nil
	settings.customBars[ID].background = nil
	table.remove(settings.customBars, ID)
end

local function removeBarByName(n)
	for i=1,#settings.customBars do
		if settings.customBars[i] and settings.customBars[i].name == n then
			settings.customBars[i].background:Hide()
			settings.customBars[i].bar = nil
			settings.customBars[i].background = nil
			table.remove(settings.customBars, i)
		end
	end
end

-- Custom bars UI

local customFrameString = "frame|TuttiHUD Custom bars frame" ..
	"|CENTER" ..
		"&text:Custom bars:200:22" ..
	"|LEFT" ..
		"&text:Colours (1-100):100:14" ..
	"|LEFT" ..
		"&text:Red:64:10" ..
		"&text:Green:64:10" ..
		"&text:Blue:64:10" ..
	"|LEFT" ..
		"&editbox:64:1" ..
		"&editbox:64:2" ..
		"&editbox:64:3" ..
	"|LEFT" ..
		"&text:Size:100:14" ..
	"|LEFT" ..
		"&text:Length:64:10" ..
		"&text:Width:64:10" ..
	"|LEFT" ..
		"&editbox:64:4" ..
		"&editbox:64:5" ..
	"|LEFT" ..
		"&checkbox:200:Make vertical" ..
	"|LEFT" ..
		"&text:Value function:100:10" ..
	"|LEFT" ..
		"&editbox:200:Six" ..
	"|LEFT" ..
		"&text:Max function:100:10" ..
	"|LEFT" ..
		"&editbox:200:Seven" ..
	"|LEFT" ..
		"&text:Bar name:100:10" ..
	"|LEFT" ..
		"&editbox:200:Eight" ..
	"|CENTER" ..
		"&button:Create" ..
		"&button:Delete" ..
	"|CENTER" ..
		"&button:Close"

local customFrame = TuttiFrameSetupParse(customFrameString)
customFrame:SetPoint("CENTER")
for i=1, 5 do customFrame.editbox[i]:SetNumeric(true) end
customFrame:Hide()

local function customFrameShow()
	for i=1, 8 do customFrame.editbox[i]:SetText("") end
	customFrame:Show()
end

local function createButton()
	local r = (tonumber(customFrame.editbox[1]:GetText()) or 0) / 100
	local g = (tonumber(customFrame.editbox[2]:GetText()) or 0) / 100
	local b = (tonumber(customFrame.editbox[3]:GetText()) or 0) / 100
	local l = tonumber(customFrame.editbox[4]:GetText())
	local w = tonumber(customFrame.editbox[5]:GetText())
	local ro = customFrame.checkbox[1]:GetChecked() and "VERTICAL" or "HORIZONTAL"
	local f = customFrame.editbox[6]:GetText()
	local m = customFrame.editbox[7]:GetText()
	local n = customFrame.editbox[8]:GetText()
	createBar(r, g, b, l, w, ro, m, f, n)
	for i=1, 8 do customFrame.editbox[i]:SetText("") end
	customFrame.checkbox[1]:SetChecked(false)
	customFrame.editbox[1]:SetFocus()
	customFrame.editbox[1]:ClearFocus()
end

local function deleteButton()
	removeBarByName(customFrame.editbox[8]:GetText())
	for i=1, 8 do customFrame.editbox[i]:SetText("") end
	customFrame.checkbox[1]:SetChecked(false)
	customFrame.editbox[1]:SetFocus()
	customFrame.editbox[1]:ClearFocus()
end

local function c_closeButton()
	customFrame:Hide()
end

customFrame.button[1]:SetScript("OnClick", createButton)
customFrame.button[2]:SetScript("OnClick", deleteButton)
customFrame.button[3]:SetScript("OnClick", c_closeButton)

-- Hook into the /tutti slash command

local function tuttiHUDSlash(arguments)
	local args = {}
	if arguments then
		arguments = arguments .. " garbage"-- Turns out the string match function is unable to return the last word with this pattern
		arg, rest = strmatch(arguments, "(.-)%s+(.*)")
		while arg do
			args[#args + 1] = arg
			arguments = rest
			arg, rest = strmatch(arguments, "(.-)%s+(.*)")
		end
	end
	if not args[1] then
		print("tuttiHUD v1.2")
		print("Available commands:")
		print("- options")
		print("- custom")
	elseif args[1] == "options" then
		showOptions()
	elseif args[1] == "custom" then
		customFrameShow()
	else
		print("Unknown command.")
		print("tuttiHUD v1.2")
		print("Available commands:")
		print("- options")
		print("- custom")
	end
end

tuttiHookSlash("hud", tuttiHUDSlash)