local function tuttiUIHelpFunc()
	print("tutti UI v0.1")
	print("Available commands:")
	for k,v in pairs(slashHooks) do
		if k ~= "" then print("- " .. k) end
	end
end

local errors = {
	["NOSUCHCOMMAND"] = "Unknown or no command. Type /tutti help for a list of available commands.",
	["BADTABLESTRING"] = "Bad table string.",
}

function tuttiUIError(ID)
	print(errors[ID] or "tuttiUI: An addon tried to call an error that does not exist. Please point the irony out to its creator.")
end

slashHooks = {
	["help"] = function() tuttiUIHelpFunc() end
}

function tuttiHookSlash(cmd, func)
	slashHooks[cmd] = func
end

function copyTable(table)
	if type(table) ~= "table" then
		return table
	else
		local newTable = {};
		for k,v in pairs(table) do
			newTable[k] = copyTable(v)
		end
		return newTable
	end
end

function tableEquals(table1, table2)
	local equals = true
	if type(table1) ~= "table" or type(table2) ~= "table" then
		if table1 == table2 then return true else return false end
	else
		for k,v in pairs(table1) do
			if not tableEquals(v, table2[k]) then equals = false end
		end
		for k,v in pairs(table2) do
			if not tableEquals(v, table1[k]) then equals = false end
		end
		return equals
	end
end

local function onEvent(frame, event, ...)
	
end

local eventFrame = CreateFrame("Frame")
eventFrame:RegisterEvent("ADDON_LOADED")
eventFrame:SetScript("OnEvent", onEvent)

-- FRAME CREATION

local tuttiBackdropStandard = {
	bgFile = "Interface\\AddOns\\tuttiUI\\textures\\grey.blp",
	edgeFile = "Interface\\AddOns\\tuttiUI\\textures\\black.blp",
	tile = true, tileSize = 8, edgeSize = 4,
	insets = { left = 0, right = 0, top = 0, bottom = 0 }
}

local tuttiBackdropBlack = {
	bgFile = "Interface\\AddOns\\tuttiUI\\textures\\black.blp",
	edgeFile = nil,
	tile = true, tileSize = 8, edgeSize = 4,
	insets = { left = 0, right = 0, top = 0, bottom = 0 }
}

local tuttiBackdropWoWStyle = {
	bgFile = "Interface\\TutorialFrame\\TutorialFrameBackground",
	edgeFile = "Interface\\DialogFrame\\UI-DialogBox-Border",
	tile = true, tileSize = 32, edgeSize = 32,
	insets = { left = 10, right = 10, top = 10, bottom = 10 }
}

local tuttiBackdropTooltip = {
	bgFile = "Interface\\TutorialFrame\\TutorialFrameBackground",
	edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border",
	tile = true, tileSize = 16, edgeSize = 16,
	insets = { left = 5, right = 5, top = 5, bottom = 5 }
}

local tuttiBackdropStandardThin = {
	bgFile = "Interface\\AddOns\\tuttiUI\\textures\\grey.blp",
	edgeFile = "Interface\\AddOns\\tuttiUI\\textures\\black.blp",
	tile = true, tileSize = 8, edgeSize = 2,
	insets = { left = 0, right = 0, top = 0, bottom = 0 }
}

local backdropList = {
	["STANDARDGREY"] = tuttiBackdropStandard,
	["ALLBLACK"] = tuttiBackdropBlack,
	["WOWSTYLE"] = tuttiBackdropWoWStyle,
	["TOOLTIP"] = tuttiBackdropTooltip,
	["STANDARDTHIN"] = tuttiBackdropStandardThin,
}

function createTuttiFrame(width, height, backdrop, hasUIParent, movable, name, parent)
	local newFrame = CreateFrame("Frame", name, hasUIParent and UIParent or parent or nil)
	newFrame:SetBackdrop(backdropList[backdrop] or backdrop)
	newFrame:SetWidth(width)
	newFrame:SetHeight(height)
	if movable then
		newFrame:EnableMouse(true)
		newFrame:SetMovable(true)
		newFrame:RegisterForDrag("LeftButton")
		newFrame:SetScript("OnDragStart", newFrame.StartMoving)
		newFrame:SetScript("OnDragStop", newFrame.StopMovingOrSizing)
	end
	return newFrame
end

function createTuttiButton(text, func, parent, name)
	local button = CreateFrame("Button", name or nil, parent, "OptionsButtonTemplate")
	button:SetText(text)
	button:SetScript("OnClick", func or nil)
	return button
end

function createTuttiLabel(frame, text, size, name)
	local label = frame:CreateFontString(name or nil)
	label:SetFont("Fonts\\FRIZQT__.TTF", size)
	label:SetText(text)
	return label
end

function createTuttiCheckbox(name, parent, text)
	local checkbox = CreateFrame("CheckButton", name, parent, "OptionsCheckButtonTemplate")
	getglobal(checkbox:GetName() .. 'Text'):SetText(text)
	return checkbox
end

function createTuttiEditbox(width, height, parent, name)
	local editbox = CreateFrame("EditBox", name or nil, parent)
	editbox:SetBackdrop(tuttiBackdropTooltip)
	editbox:SetWidth(width)
	editbox:SetHeight(height)
	editbox:SetAutoFocus(false)
	editbox:SetScript("OnEscapePressed", editbox.ClearFocus)
	editbox:SetScript("OnEnterPressed", editbox.ClearFocus)
	editbox:SetFont("Fonts\\FRIZQT__.TTF", 12)
	editbox:SetTextInsets(6,0,0,0)
	return editbox
end

function changeTuttiBackdrop(frame, backdrop)
	frame:SetBackdrop(backdropList[backdrop] or backdrop)
end

SLASH_TUTTIUI1 = '/tutti'

local function tuttiSlash(args, editbox)
	if args then
		local command, arguments = args:match("(.-)%s+(.*)")
		if not command then command = args end
		if slashHooks[command] then slashHooks[command](arguments) else tuttiUIError("NOSUCHCOMMAND") end
	end
end

SlashCmdList["TUTTIUI"] = tuttiSlash;

--------------------------
-- Advanced frame setup --
--------------------------

-- Use:	Pass all information on all widgets the frame should have.
--      The frame will be set up accordingly.
-- Returns: The frame that was created. Arrays holding the widgets
--          will be created as children of the frame.
-- Specifics:

-- frameTable = {
--	name = "name",
--	{ line 1 },
--	{ line 2 },
--	...
-- }

-- line = {
--	alignment = "LEFT"/"CENTER"/"RIGHT",
--	{ widget 1 },
--	{ widget 2 },
--	...
-- }

-- widget = {
--	type = "type",
--	parameter1 = <p>,
--	parameter2 = <p>,
--	...
-- }

-- types:
--	"space"; pushes elements below further down, and elements to the right further right.
--		- width; the height of the empty space
--		- height; the height of the empty space
--	"text"
--		- text; the text to display
--		- width; the width reserved for the string
--		- size; the size of the text
--		- name; (optional) the name of the label object
--	"editbox"
--		- width
--		- text
--		- name; (optional) the name of the editbox
--	"checkbox"
--		- width; the width reserved for the checkbox
--		- text; the label to the right of the checkbox
--		- name; (optional) the name of the checkbox
--	"button"; a standard, red button
--		- text
--		- name; (optional) the name of the button

-- Names: If a name is provided, the frame can later be referred to by that name.
--        An unnamed widget has to be referred through the widget arrays, which
--        can cause problems if the frame is later edited and the widget order is
--        shifted.

local function placeWidget(frame, widget, align, x, y, lineWidth, width)
	local point = "TOP"
	if align ~= "CENTER" then point = point .. align end
	local xPos = x
	if align == "LEFT" then xPos = xPos + 12
	elseif align == "CENTER" then xPos = xPos - (lineWidth / 2) + (width / 2)
	elseif align == "RIGHT" then xPos = xPos - lineWidth + width - 12 end
	widget:SetPoint(point, frame, point, xPos, -y - 12)
end

function TuttiFrameSetup(frameTable)
	local f = createTuttiFrame(24, 24, "WOWSTYLE", true, true, frameTable.name)
	local frameHeight = 24
	local frameWidth = 24
	local curYPos = 0
	local text = {}
	local editbox = {}
	local checkbox = {}
	local button = {}
	for i=1,#frameTable do -- for each line
		local lineWidth = -4
		local curXPos = 0
		local maxHeight = 0
		for j=1,#frameTable[i] do -- calculate the line's width
			lineWidth = lineWidth + 4 + (frameTable[i][j].width or ((frameTable[i][j].type == "button") and 90))
			frameWidth = max(frameWidth, lineWidth + 24)
		end
		for j=1,#frameTable[i] do -- for each widget in the line
			if frameTable[i][j].type == "space" then
				maxHeight = max(maxHeight, frameTable[i][j].height)
				curXPos = curXPos + frameTable[i][j].width + 4
			elseif frameTable[i][j].type == "text" then
				maxHeight = max(maxHeight, frameTable[i][j].size)
				local w = createTuttiLabel(f, frameTable[i][j].text, frameTable[i][j].size, frameTable[i][j].name or nil)
				placeWidget(f, w, frameTable[i].alignment, curXPos, curYPos, lineWidth, frameTable[i][j].width)
				curXPos = curXPos + frameTable[i][j].width + 4
				text[#text + 1] = w
			elseif frameTable[i][j].type == "editbox" then
				maxHeight = max(maxHeight, 24)
				local w = createTuttiEditbox(frameTable[i][j].width, 24, f, frameTable[i][j].name or nil)
				w:SetText(frameTable[i][j].text)
				placeWidget(f, w, frameTable[i].alignment, curXPos, curYPos, lineWidth, frameTable[i][j].width)
				curXPos = curXPos + frameTable[i][j].width + 4
				editbox[#editbox + 1] = w
			elseif frameTable[i][j].type == "checkbox" then
				maxHeight = max(maxHeight, 18)
				local w = createTuttiCheckbox(frameTable.name .. " " .. i .. "-" .. j, f, frameTable[i][j].text, frameTable[i][j].name or nil)
				placeWidget(f, w, frameTable[i].alignment, curXPos, curYPos-4, lineWidth, frameTable[i][j].width)
				curXPos = curXPos + frameTable[i][j].width + 4
				checkbox[#checkbox + 1] = w
			elseif frameTable[i][j].type == "button" then
				maxHeight = max(maxHeight, 18)
				local w = createTuttiButton(frameTable[i][j].text, nil, f, frameTable[i][j].name or nil)
				placeWidget(f, w, frameTable[i].alignment, curXPos, curYPos, lineWidth, 90)
				curXPos = curXPos + 94
				button[#button + 1] = w
				
			end
		end
		curYPos = curYPos + maxHeight + 4
		frameHeight = frameHeight + maxHeight + 4
	end
	f:SetWidth(frameWidth)
	f:SetHeight(frameHeight)
	f.text = text
	f.editbox = editbox
	f.checkbox = checkbox
	f.button = button
	return f
end

-- Use:	Set up a string containing the frame info
--	using the below syntax:

-- "|" splits lines
-- "&" splits widgets
-- ":" splits parameters

-- The first 'line' must be "frame"
-- The second 'line' is the frame's name
-- The third 'line' is the first actual line, and so on

-- Similarly, the first 'widget' is the line's alignment,
-- and the first 'parameter' is the widget's type.

-- Names: To name a widget, append it to the definition as an additional parameter.
-- This is optional.

function TuttiFrameSetupParse(frameString)
	-- Parses a string into a frameTable, then creates a frame based on that and returns it.
	-- Syntax:
	-- ":" divides parameters
	-- "&" divides widgets
	-- "|" divides lines

	-- Split the initial string into lines
	local lines = {}
	for i in string.gmatch(frameString, "[^|]+") do
		lines[#lines+1] = i
	end
	if lines[1] == "frame" and #lines > 2 then
		local frameTable = { name = tostring(lines[2]) }
		for i=3,#lines do
			frameTable[i-2] = {}
			local widgets = {}
			-- Split each line into widgets
			for j in string.gmatch(lines[i], "[^&]+") do
				frameTable[i-2][#frameTable[i-2]+1] = {}
				widgets[#widgets+1] = j
				local params = {}
				-- Split each widget into parameters
				for k in string.gmatch(widgets[#widgets], "[^:]+") do
					params[#params+1] = k
				end
				frameTable[i-2][#frameTable[i-2]].type = params[1]
				if params[1] == "space" then
					frameTable[i-2][#frameTable[i-2]].width = tonumber(params[2])
					frameTable[i-2][#frameTable[i-2]].height = tonumber(params[3])
				elseif params[1] == "text" then
					frameTable[i-2][#frameTable[i-2]].text = params[2]
					frameTable[i-2][#frameTable[i-2]].width = tonumber(params[3])
					frameTable[i-2][#frameTable[i-2]].size = tonumber(params[4])
					if params[5] then frameTable[i-2][#frameTable[i-2]].name = params[5] end
				elseif params[1] == "editbox" then
					frameTable[i-2][#frameTable[i-2]].width = tonumber(params[2])
					frameTable[i-2][#frameTable[i-2]].text = params[3]
					if params[4] then frameTable[i-2][#frameTable[i-2]].name = params[4] end
				elseif params[1] == "checkbox" then
					frameTable[i-2][#frameTable[i-2]].width = tonumber(params[2])
					frameTable[i-2][#frameTable[i-2]].text = params[3]
					if params[4] then frameTable[i-2][#frameTable[i-2]].name = params[4] end
				elseif params[1] == "button" then
					frameTable[i-2][#frameTable[i-2]].text = params[2]
					if params[3] then frameTable[i-2][#frameTable[i-2]].name = params[3] end
				end
			end
			tremove(frameTable[i-2], 1)
			frameTable[i-2].alignment = widgets[1]
		end
		local frame = TuttiFrameSetup(frameTable)
		return frame
	else
		tuttiUIError("BADTABLESTRING")
		local frame = CreateFrame("Frame")
		return frame
	end
end

--------------------------------
-- The following table        --
-- and the following string,  --
-- when parsed to their       --
-- respective functions, will --
-- produce identical frames   --
-- (except their names).      --
--------------------------------

--[[
local testFrameTable = {
	name = "Test Frame",
	{
		alignment = "LEFT",
		{
			type = "text",
			text = "Testing",
			width = 50,
			size = 10,
		},
		{
			type = "text",
			text = "And another",
			width = 100,
			size = 10,
		},
	},
	{
		alignment = "LEFT",
		{
			type = "text",
			text = "Line 2",
			width = 50,
			size = 10,
		}
	},
	{
		alignment = "LEFT",
		{
			type = "text",
			text = "Line 3",
			width = 50,
			size = 10,
		},
	},
	{
		alignment = "LEFT",
		{
			type = "editbox",
			width = 64,
			text = "Box 1",
		},
		{
			type = "editbox",
			width = 64,
			text = "Box 2",
		},
		{
			type = "editbox",
			width = 64,
			text = "Box 3",
		},
	},
	{
		alignment = "LEFT",
		{
			type = "editbox",
			width = 64,
			text = "Box 4",
		},
		{
			type = "editbox",
			width = 64,
			text = "Box 5",
		},
		{
			type = "editbox",
			width = 64,
			text = "Box 6",
		},
	},
	{
		alignment = "LEFT",
		{
			type = "checkbox",
			width = 100,
			text = "This is a checkbox."
		},
	},
	{
		alignment = "LEFT",
		{
			type = "checkbox",
			width = 100,
			text = "This is another checkbox."
		},
	},
	{
		alignment = "CENTER",
		{
			type = "button",
			text = "Button 1",
			name = "Button1Name",
		},
		{
			type = "button",
			text = "Button 2",
		},
	},
}

local testFrameString = "frame|Test frame 2" ..
	"|LEFT" ..
		"&text:Testing:50:10" ..
		"&text:And another:100:10" ..
	"|LEFT" ..
		"&text:Line 2:50:10" ..
	"|LEFT" ..
		"&text:Line 3:50:10" ..
	"|LEFT" ..
		"&editbox:64:Box 1" ..
		"&editbox:64:Box 2" ..
		"&editbox:64:Box 3" ..
	"|LEFT" ..
		"&editbox:64:Box 4" ..
		"&editbox:64:Box 5" ..
		"&editbox:64:Box 6" ..
	"|LEFT" ..
		"&checkbox:100:This is a checkbox." ..
	"|LEFT" ..
		"&checkbox:100:This is another checkbox." ..
	"|CENTER" ..
		"&button:Button 1:Button1Name" ..
		"&button:Button 2"

]]